from django.contrib.auth.models import User
from django.core.management import call_command
from django.core.management.base import CommandError
from django.test import override_settings, TestCase
from uniauth.models import LinkedEmail, Institution, UserProfile
import os
import sys
try:
    import mock
except ImportError:
    from unittest import mock


class AddInstitutionCommandTests(TestCase):
    """
    Tests the add_institution management command
    """

    def setUp(self):
        Institution.objects.all().delete()
        sys.stdout = open(os.devnull, 'w')

    def _check_institutions(self, actual, expected):
        act_values = [(x.name, x.slug, x.cas_server_url) for x in actual]
        exp_values = [(x.name, x.slug, x.cas_server_url) for x in expected]
        self.assertEqual(act_values, exp_values)

    def test_add_institution_command_valid_inputs(self):
        """
        Ensures command works as expected for valid inputs
        """
        call_command("add_institution", "Test", "https://www.example.com")
        call_command("add_institution", "Other Inst", "http://fed.foobar.edu/")
        call_command("add_institution", "Test", "https://www.example.com",
            "--update-existing")
        actual = Institution.objects.order_by("slug")
        expected = [
                Institution(name="Other Inst", slug="other-inst",
                    cas_server_url="http://fed.foobar.edu/"),
                Institution(name="Test", slug="test",
                    cas_server_url="https://www.example.com"),
        ]
        self._check_institutions(actual, expected)

    def test_add_institution_command_update_existing_valid_inputs(self):
        """
        Ensures command works as expected for valid inputs with the
        --update-existing option.
        """
        call_command("add_institution", "Test", "https://www.example.com",
            "--update-existing")
        call_command("add_institution", "Other Inst", "http://fed.foobar.edu/")
        call_command("add_institution", "Test", "https://fed.university.edu/",
             "--update-existing")
        actual = Institution.objects.order_by("slug")
        expected = [
                Institution(name="Other Inst", slug="other-inst",
                    cas_server_url="http://fed.foobar.edu/"),
                Institution(name="Test", slug="test",
                    cas_server_url="https://fed.university.edu/"),
        ]
        self._check_institutions(actual, expected)

    def test_add_institution_command_invalid_inputs(self):
        """
        Ensures command fails gracefully for invalid inputs
        """
        self.assertRaisesRegex(CommandError, "argument", call_command,
                "add_institution")
        self.assertRaisesRegex(CommandError, "argument", call_command,
                "add_institution", "Some Name")
        self.assertRaisesRegex(CommandError, "malformed", call_command,
                "add_institution", "", "")
        self.assertRaisesRegex(CommandError, "malformed", call_command,
                "add_institution", "name", "https://www.example.x")
        call_command("add_institution", "Test", "https://www.example.com")
        self.assertRaisesRegex(CommandError, "exists", call_command,
                "add_institution", "test", "https://www.foo.bar")


class FlushTmpUsersTests(TestCase):
    """
    Tests the flush_tmp_users management command
    """

    @mock.patch("uniauth.management.commands.flush_tmp_users.get_input")
    @mock.patch("uniauth.management.commands.flush_tmp_users.flush_old_tmp_users")
    def test_flush_tmp_users_command_correct(self, mock_flush, mock_get_input):
        """
        Ensure command works as expected given valid starting conditions
        Ensures command propagates optional arguments properly, and uses
        expected default values when optional arguments are not provided
        """
        # Ensure command fails gracefully with invalid arguments
        self.assertRaisesRegex(CommandError, "days", call_command,
                "flush_tmp_users", "dne")

        # Ensure nothing happens when the user does not agree to continue
        mock_get_input.return_value = "no"
        call_command("flush_tmp_users")
        mock_flush.assert_not_called()

        # Ensure command propagates the optional days argument properly
        mock_get_input.return_value = "yes"
        call_command("flush_tmp_users", 4)
        mock_flush.assert_called_once_with(days=4)

        # Ensure command propagates uses expected default otherwise
        call_command("flush_tmp_users")
        mock_flush.assert_called_with(days=1)


class MigrateCASCommandTests(TestCase):
    """
    Tests the migrate_cas management command
    """

    def setUp(self):
        User.objects.all().delete()
        self.ex = User.objects.create(username="exid123")
        self.john = User.objects.create(username="johndoe")
        self.mary = User.objects.create(username="marysue")
        self.adam = User.objects.create(username="adam998")
        UserProfile.objects.all().delete()
        UserProfile.objects.create(user=self.adam)
        Institution.objects.create(name="Example Inst", slug="example-inst",
                cas_server_url="https://fake.example.edu")

    @mock.patch("uniauth.management.commands.migrate_cas.get_input")
    def test_migrate_cas_command_correct(self, mock_get_input):
        """
        Ensures command works as expected given valid starting conditions
        """
        # Ensure command fails gracefully with invalid arguments
        self.assertRaisesRegex(CommandError, "argument", call_command,
                "migrate_cas")
        self.assertRaisesRegex(CommandError, "slug", call_command,
                "migrate_cas", "dne")
        # Ensure nothing happens when the user does not agree to continue
        mock_get_input.return_value = "no"
        call_command("migrate_cas", "example-inst")
        self.assertEqual(UserProfile.objects.count(), 1)
        mock_get_input.return_value = "abcde"
        call_command("migrate_cas", "example-inst")
        self.assertEqual(UserProfile.objects.count(), 1)
        # Ensure migration occurs when user does agree to continue
        mock_get_input.return_value = "yes"
        call_command("migrate_cas", "example-inst")
        self.assertEqual(UserProfile.objects.count(), 4)
        actual_usernames = User.objects.values_list("username", flat=True)
        expected_usernames = ["adam998", "cas-example-inst-exid123",
                "cas-example-inst-johndoe", "cas-example-inst-marysue"]
        self.assertEqual(sorted(actual_usernames), expected_usernames)


class MigrateCustomCommandTests(TestCase):
    """
    Tests the migrate_custom management command
    """

    def setUp(self):
        User.objects.all().delete()
        self.ex = User.objects.create(username="exid123")
        self.john = User.objects.create(username="johndoe",
                password="johnpass")
        self.mary = User.objects.create(username="marysue",
                email="mary.sue@gmail.com", password="marypass")
        self.adam = User.objects.create(username="adam998@example.com",
                email="adam998@example.com", password="adampass")
        UserProfile.objects.all().delete()
        UserProfile.objects.create(user=self.adam)
        LinkedEmail.objects.all().delete()

    @mock.patch("uniauth.management.commands.migrate_custom.get_input")
    def test_migrate_custom_command_correct(self, mock_get_input):
        """
        Ensures command works as expected given valid starting conditions
        """
        # Ensure nothing happens when the user does not agree to continue
        mock_get_input.return_value = "no"
        call_command("migrate_custom")
        self.assertEqual(UserProfile.objects.count(), 1)
        mock_get_input.return_value = "abcde"
        call_command("migrate_custom")
        self.assertEqual(UserProfile.objects.count(), 1)
        # Ensure migration occurs when user does agree to continue
        mock_get_input.return_value = "yes"
        call_command("migrate_custom")
        self.assertEqual(UserProfile.objects.count(), 3)
        self.assertTrue(UserProfile.objects.filter(user__username="johndoe")\
                .exists())
        self.assertEqual(LinkedEmail.objects.count(), 1)
        self.assertTrue(LinkedEmail.objects.filter(address="mary.sue@gmail.com",
                profile__user__username="marysue").exists())
        self.assertEqual(self.john.uniauth_profile.linked_emails.count(), 0)


class RemoveInsitutionCommandTests(TestCase):
    """
    Tests the remove_institution management command
    """

    def setUp(self):
        Institution.objects.all().delete()
        Institution.objects.create(name="Harvard", slug="harvard",
                cas_server_url="https://fake.harvardcas.edu/")
        Institution.objects.create(name="Yale", slug="yale",
                cas_server_url="https://fake.yalecas.edu/")
        Institution.objects.create(name="Penn State", slug="penn-state",
                cas_server_url="https://fake.penncas.edu/")
        sys.stdout = open(os.devnull, 'w')

    def _check_institutions(self, actual, expected):
        act_values = [(x.name, x.slug, x.cas_server_url) for x in actual]
        exp_values = [(x.name, x.slug, x.cas_server_url) for x in expected]
        self.assertEqual(act_values, exp_values)

    @mock.patch("uniauth.management.commands.remove_institution.get_input")
    def test_remove_institution_command_valid_inputs(self, mock_get_input):
        """
        Ensures command works as expected for valid inputs
        """
        mock_get_input.return_value = "yes"
        call_command("remove_institution", "penn-state")
        mock_get_input.return_value = "y"
        call_command("remove_institution", "harvard")
        mock_get_input.return_value = "n"
        call_command("remove_institution", "yale")
        mock_get_input.return_value = "other"
        call_command("remove_institution", "yale")
        actual = Institution.objects.order_by("slug")
        expected = [Institution(name="Yale", slug="yale",
                cas_server_url="https://fake.yalecas.edu/")]
        self._check_institutions(actual, expected)

    @mock.patch("uniauth.management.commands.remove_institution.get_input")
    def test_remove_institution_command_invalid_inputs(self, mock_get_input):
        """
        Ensures command fails gracefully for invalid inputs
        """
        self.assertRaisesRegex(CommandError, "argument", call_command,
                "remove_institution")
        self.assertRaisesRegex(CommandError, "slug", call_command,
                "remove_institution", "dne")
        mock_get_input.return_value = "yes"
        call_command("remove_institution", "yale")
        self.assertRaisesRegex(CommandError, "exists", call_command,
                "remove_institution", "yale")
        actual = Institution.objects.order_by("slug")
        expected = [
                Institution(name="Harvard", slug="harvard",
                        cas_server_url="https://fake.harvardcas.edu/"),
                Institution(name="Penn State", slug="penn-state",
                        cas_server_url="https://fake.penncas.edu/")
        ]
        self._check_institutions(actual, expected)




from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import AuthenticationForm, \
        PasswordChangeForm as AuthPasswordChangeForm, \
        PasswordResetForm as AuthPasswordResetForm, \
        SetPasswordForm as AuthSetPasswordForm, UserCreationForm
try:
    from django.utils.translation import ugettext_lazy as _
except ImportError:
    from django.utils.translation import gettext_lazy as _
from uniauth.models import LinkedEmail
from uniauth.utils import get_setting


class AddLinkedEmailForm(forms.Form):
    """
    Form for adding a linked email address to a profile.
    """

    email = forms.EmailField(max_length=254, label="Email address")

    def __init__(self, user, *args, **kwargs):
        super(AddLinkedEmailForm, self).__init__(*args, **kwargs)
        self.user = user

    def clean_email(self):
        """
        Ensures you can't link an email that has already
        been linked to the current Uniauth profile.

        If UNIAUTH_ALLOW_SHARED_EMAILS is False, ensures
        the email hasn't been linked to any profile.
        """
        email = self.cleaned_data.get('email')
        if LinkedEmail.objects.filter(profile=self.user.uniauth_profile,
                address=email).exists():
            err_msg = ("That email address has already been linked "
                    "to this account.")
            raise forms.ValidationError(err_msg, code="already_linked")
        if not get_setting('UNIAUTH_ALLOW_SHARED_EMAILS') and \
                LinkedEmail.objects.filter(address=email,
                        is_verified=True).exists():
            err_msg = ("That email address has already been linked "
                    "to another account.")
            raise forms.ValidationError(err_msg, code="already_linked")
        return email

    def clean(self):
        """
        Ensure the user does not link more than the
        maximum number of linked emails per user.
        """
        cleaned_data = super(AddLinkedEmailForm, self).clean()
        max_linked_emails = get_setting('UNIAUTH_MAX_LINKED_EMAILS')
        num_linked_emails = self.user.uniauth_profile.linked_emails.count()
        if max_linked_emails > 0 and num_linked_emails >= max_linked_emails:
            err_msg = ("You can not link more than %d emails to your account."
                    % max_linked_emails)
            raise forms.ValidationError(err_msg, code="max_emails")
        return cleaned_data


class ChangePrimaryEmailForm(forms.Form):
    """
    Form for changing a user's primary email address.
    """

    def __init__(self, user, *args, **kwargs):
        """
        Set the choices to the current profile's verified linked emails.
        """
        super(ChangePrimaryEmailForm, self).__init__(*args, **kwargs)
        self.user = user
        verified_emails = LinkedEmail.objects.filter(
                profile=self.user.uniauth_profile, is_verified=True).all()
        choices = map(lambda x: (x.address, x.address), verified_emails)
        self.fields['email'] = forms.ChoiceField(choices=choices)
        self.fields['email'].initial = self.user.email

    def clean_email(self):
        """
        Ensures you can't set your primary email address to one
        another verified user already has their primary email.
        """
        email = self.cleaned_data.get('email')
        if get_user_model().objects.filter(email=email)\
                .exclude(pk=self.user.pk).exists():
            err_msg = ("A user with that primary email address already "
                    "exists. Please choose another.")
            raise forms.ValidationError(err_msg, code="email_taken")
        return email


class LinkedEmailActionForm(forms.Form):
    """
    Form for deleting linked emails or resending
    verification emails for pending linked emails.
    """
    delete_pk = forms.IntegerField(required=False)
    resend_pk = forms.IntegerField(required=False)


class LoginForm(AuthenticationForm):
    """
    Form for logging via Uniauth credentials.

    Subclasses AuthenticationForm, which has username
    and password fields + some additional conveniences.

    Note: The email address field is left as 'username'
    so we can leverage the built-in methods that depend
    on it being named 'username'.
    """

    def __init__(self, request=None, *args, **kwargs):
        """
        Change the invalid login error message to mention
        using a correct email address instead of username.
        """
        super(LoginForm, self).__init__(request, *args, **kwargs)
        self.error_messages['invalid_login'] = _(
                "Please enter a correct email and password."
        )


def _prevent_shared_email_and_password(linked_emails, new_password):
    """
    Ensures the proposed new password is different from all
    passwords used by users that have a linked email in the
    provided list.
    """
    users_with_shared_email = set()

    # Get the set of users who share an email with this user
    for linked_email in linked_emails:
        users = get_user_model().objects.filter(
                uniauth_profile__linked_emails__address__iexact=linked_email,
                uniauth_profile__linked_emails__is_verified=True,
                is_active=True
        ).all()
        users_with_shared_email.update(users)

    # Make sure the new password doesn't match any of theirs
    for user in users_with_shared_email:
        if user.check_password(new_password):
            err_msg = "Please choose a different password."
            raise forms.ValidationError(err_msg, code="password_taken")

    return new_password


class SetPasswordForm(AuthSetPasswordForm):
    """
    Form allowing a user to change their password without
    entering their old one.

    Subclasses the built-in SetPasswordForm, which does
    most of the heavy lifting + has several security features.
    """

    def clean_new_password1(self):
        """
        Prevent users from setting their password to the same
        password another user with a shared linked email is using.
        """
        new_password = self.cleaned_data.get('new_password1')
        linked_emails = self.user.uniauth_profile.linked_emails\
                .values_list('address', flat=True)
        _prevent_shared_email_and_password(linked_emails, new_password)
        return new_password


class PasswordChangeForm(AuthPasswordChangeForm):
    """
    Form allowing a user to change their password by entering
    their old one first.

    Subclasses the built-in PasswordChangeForm, which does
    most of the heavy lifting + has several security features.
    """

    def clean_new_password1(self):
        """
        Prevent users from setting their password to the same
        password another user with a shared linked email is using.
        """
        new_password = self.cleaned_data.get('new_password1')
        linked_emails = self.user.uniauth_profile.linked_emails\
                .values_list('address', flat=True)
        _prevent_shared_email_and_password(linked_emails, new_password)
        return new_password


class PasswordResetForm(AuthPasswordResetForm):
    """
    Form for asking for a password reset.

    Subclasses the built-in PasswordResetForm, which does
    most of the heavy lifting + has several security features.
    """

    def get_users(self, email):
        """
        Given an email, returns matching user(s) who should
        receive a password reset link.

        Modified to find all users who possess the provided
        address as a verified linked email instead of just
        primary email.
        """
        users = get_user_model().objects.filter(
                uniauth_profile__linked_emails__address__iexact=email,
                uniauth_profile__linked_emails__is_verified=True,
                is_active=True
        )
        return (u for u in users if u.has_usable_password())


class SignupForm(UserCreationForm):
    """
    Form for creating a new Uniauth account.

    Subclasses UserCreationForm, which has two password fields
    (one is for confirming it) + some additional conveniences.
    """

    email = forms.EmailField(max_length=254, label="Email address",
            help_text="Primary email address.")

    class Meta:
        model = get_user_model()
        fields = ('email', 'password1', 'password2')

    def clean_email(self):
        """
        Ensures that you can't sign up with an email address
        another verified user already has their primary email.

        If UNIAUTH_ALLOW_SHARED_EMAILS is False, ensures the
        email hasn't been linked + verified by any profile.
        """
        email = self.cleaned_data.get('email')
        if get_user_model().objects.filter(email=email).exists():
            err_msg = ("A user with that primary email address already "
                    "exists. Please choose another.")
            raise forms.ValidationError(err_msg, code="email_taken")
        if not get_setting('UNIAUTH_ALLOW_SHARED_EMAILS') and \
                LinkedEmail.objects.filter(address=email,
                        is_verified=True).exists():
            err_msg = ("That email address has already been linked "
                    "to another account.")
            raise forms.ValidationError(err_msg, code="already_linked")
        return email

    def clean_password1(self):
        """
        Prevent users from setting their password to the same
        password another user with a shared linked email is using.
        """
        email = self.cleaned_data.get('email')
        password = self.cleaned_data.get('password1')
        _prevent_shared_email_and_password([email], password)
        return password




from django.contrib.auth.tokens import PasswordResetTokenGenerator
from rest_framework_simplejwt.tokens import RefreshToken


class EmailVerificationTokenGenerator(PasswordResetTokenGenerator):
    """
    Creates tokens for linked email verification.
    """

    def _make_hash_value(self, email, timestamp):
        return str(email.pk) + str(email.is_verified) + str(timestamp)


token_generator = EmailVerificationTokenGenerator()


def get_jwt_tokens_for_user(user, **kwargs):
    """
    Generates a refresh token for the valid user
    """
    refresh = RefreshToken.for_user(user)

    return str(refresh), str(refresh.access_token)


# Generated by Django 3.1.5 on 2022-11-07 23:53

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('uniauth', '0002_auto_20190707_2305'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='institutionaccount',
            unique_together={('institution', 'cas_id')},
        ),
    ]



try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url
from django.core.exceptions import ImproperlyConfigured
from uniauth import views
from uniauth.utils import get_setting

# The standard login form requires URLs that don't exist when
# cas_only is used. Ensure that this setting is False.
if get_setting('UNIAUTH_LOGIN_DISPLAY_STANDARD'):
    err_msg = ("'uniauth.urls.cas_only' can not be used when %s is True. "
            "Please include 'uniauth.urls' instead.") % \
            'UNIAUTH_LOGIN_DISPLAY_STANDARD'
    raise ImproperlyConfigured(err_msg)

app_name = 'uniauth'

urlpatterns = [
    url(r'^login/$', views.login, name='login'),
    url(r'^cas-login/(?P<institution>[a-z0-9\-]+)/$', views.cas_login, name='cas-login'),
    url(r'^logout/$', views.logout, name='logout'),
]



from django.contrib.auth.models import User
from django.test import override_settings, RequestFactory, TestCase
from uniauth.backends import CASBackend, LinkedEmailBackend, \
        UsernameOrLinkedEmailBackend
from uniauth.models import Institution, InstitutionAccount, LinkedEmail
try:
    import mock
except ImportError:
    from unittest import mock


class CASBackendTests(TestCase):
    """
    Tests the CASBackend in backends.py
    """

    def setUp(self):
        self.factory = RequestFactory()
        self.inst = Institution.objects.create(name="Test Inst",
                slug="test-inst", cas_server_url="https://fed.testinst.edu/")
        self.inst2 = Institution.objects.create(name="Other Inst",
                slug="other-inst", cas_server_url="https://fed.other.edu/")

    def _get_request(self):
        request = self.factory.get("/accounts/cas-login/")
        request.session = {}
        return request

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_verification_failed(self, mock_verify_ticket):
        """
        Ensure backend returns None if the ticket fails to verify
        """
        mock_verify_ticket.return_value = (None, {}, None)
        backend = CASBackend()
        request = self._get_request()

        prev_num_users = User.objects.count()
        user = backend.authenticate(request, institution=self.inst,
                ticket="fake-ticket", service="http://www.service.com/")
        self.assertEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users)

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_create_user(self, mock_verify_ticket):
        """
        Ensure backend properly creates a new user when necessary
        """
        mock_verify_ticket.return_value = ("newuser", {"ticket": "fake-ticket",
            "service": "http://www.service.com/"}, None)
        backend = CASBackend()
        request = self._get_request()

        self.assertFalse(User.objects.filter(username="newuser").exists())
        prev_num_users = User.objects.count()
        user = backend.authenticate(request, institution=self.inst,
                ticket="fake-ticket", service="http://www.service.com/")
        self.assertNotEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users+1)
        self.assertTrue(User.objects.filter(username="cas-test-inst-newuser")\
                .exists())
        self.assertEqual(request.session["attributes"],
                {"ticket": "fake-ticket", "service": "http://www.service.com/"})

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_existing_user(self, mock_verify_ticket):
        """
        Ensure backend returns existing user if possible
        """
        backend = CASBackend()
        user1 = User.objects.create(username="johndoe@gmail.com",
                email="john.doe@aol.com")
        InstitutionAccount.objects.create(profile=user1.uniauth_profile,
                institution=self.inst, cas_id="john123")
        user2 = User.objects.create(username="cas-test-inst-jane987")
        fakeout = User.objects.create(username="fakeout@gmail.com",
                email="fakeout@verizon.net")
        InstitutionAccount.objects.create(profile=fakeout.uniauth_profile,
                institution=self.inst2, cas_id="john123")
        prev_num_users = User.objects.count()

        # Test case where there is a user with an InstitutionAccount
        mock_verify_ticket.return_value = ("john123", {"ticket": "fake-ticket",
            "service": "http://www.example.com/"}, None)
        request = self._get_request()
        user = backend.authenticate(request, institution=self.inst,
                ticket="fake-ticket", service="http://www.example.com/")
        self.assertNotEqual(user, None)
        self.assertEqual(user.username, "johndoe@gmail.com")
        self.assertEqual(User.objects.count(), prev_num_users)
        self.assertEqual(request.session["attributes"],
                {"ticket": "fake-ticket", "service": "http://www.example.com/"})

        # Test case where there is an existing unlinked account user
        mock_verify_ticket.return_value = ("jane987", {"ticket": "ticket0123",
            "service": "http://someservice.gov/"}, None)
        request = self._get_request()
        user = backend.authenticate(request, institution=self.inst,
                ticket="ticket0123", service="http://someservice.gov/")
        self.assertNotEqual(user, None)
        self.assertEqual(user.username, "cas-test-inst-jane987")
        self.assertEqual(User.objects.count(), prev_num_users)
        self.assertEqual(request.session["attributes"],
                {"ticket": "ticket0123", "service": "http://someservice.gov/"})

    @mock.patch("cas.CASClientV2.verify_ticket")
    def test_cas_backend_missing_request(self, mock_verify_ticket):
        """
        Ensure authentication still works without the request parameter
        """
        backend = CASBackend()

        # Test creating a new user
        mock_verify_ticket.return_value = ("newuser", {"ticket": "fake-ticket",
            "service": "http://www.service.com/"}, None)
        self.assertFalse(User.objects.filter(username="newuser").exists())
        prev_num_users = User.objects.count()
        user = backend.authenticate(None, institution=self.inst,
                ticket="fake-ticket", service="http://www.service.com/")
        self.assertNotEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users+1)
        self.assertTrue(User.objects.filter(username="cas-test-inst-newuser")\
                .exists())

        # Test finding an existing user
        mock_verify_ticket.return_value = ("jane987", {"ticket": "ticket0123",
            "service": "http://someservice.gov/"}, None)
        User.objects.create(username="cas-test-inst-jane987")
        prev_num_users = User.objects.count()
        user = backend.authenticate(None, institution=self.inst,
                ticket="ticket0123", service="http://someservice.gov/")
        self.assertNotEqual(user, None)
        self.assertEqual(user.username, "cas-test-inst-jane987")
        self.assertEqual(User.objects.count(), prev_num_users)

        # Test failed authentication
        mock_verify_ticket.return_value = (None, {}, None)
        prev_num_users = User.objects.count()
        user = backend.authenticate(None, institution=self.inst,
                ticket="fake-ticket", service="http://www.service.com/")
        self.assertEqual(user, None)
        self.assertEqual(User.objects.count(), prev_num_users)


class EmailBackendTests(TestCase):
    """
    Parent class for the *EmailBackendTests
    """

    def setUp(self):
        self.factory = RequestFactory()
        self.john = User.objects.create_user(username="johndoe@gmail.com",
                email="johndoe@gmail.com", password="johnpass")
        self.mary = User.objects.create_user(username="marysue@outlook.com",
                email="mary.sue@gmail.com", password="marypass")
        LinkedEmail.objects.create(profile=self.mary.uniauth_profile,
                address="alternate@gmail.com", is_verified=True)
        LinkedEmail.objects.create(profile=self.mary.uniauth_profile,
                address="pending@gmail.com", is_verified=False)
        self.cas = User.objects.create_user(username="cas-inst-netid123")
        self.tmp = User.objects.create_user(username="tmp-0123_456",
                password="tmppass")
        self.perm = User.objects.create_user(username="permid555",
                password="permpass")
        self.mr = User.objects.create_user(username="mrjones@gmail.com",
                email="mrjones@gmail.com", password="mrpass")
        self.ms = User.objects.create_user(username="msjones@gmail.com",
                email="msjones@gmail.com", password="mspass")
        LinkedEmail.objects.create(profile=self.mr.uniauth_profile,
                address="sharedjones@gmail.com", is_verified=True)
        LinkedEmail.objects.create(profile=self.ms.uniauth_profile,
                address="sharedjones@gmail.com", is_verified=True)


@override_settings(UNIAUTH_ALLOW_SHARED_EMAILS=True)
class LinkedEmailBackendTests(EmailBackendTests):
    """
    Tests the LinkedEmailBackend in backends.py
    """

    def test_linked_email_backend_valid_credentials(self):
        """
        Ensure backend returns appropriate user when valid
        credentials are provided
        """
        backend = LinkedEmailBackend()

        # Log in with email + password
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password="johnpass")
        self.assertEqual(user, self.john)

        # Log in with email (under "username" parameter) + password
        user = backend.authenticate(None, username="mary.sue@gmail.com",
                password="marypass")
        self.assertEqual(user, self.mary)

        # Log in with verified linked email
        user = backend.authenticate(None, email="alternate@gmail.com",
                password="marypass")
        self.assertEqual(user, self.mary)

        # Log in with shared email address
        user = backend.authenticate(None, email="sharedjones@gmail.com",
                password="mspass")
        self.assertEqual(user, self.ms)
        user = backend.authenticate(None, email="sharedjones@gmail.com",
                password="mrpass")
        self.assertEqual(user, self.mr)

    def test_linked_email_backend_invalid_credentials(self):
        """
        Ensure authentication fails when invalid credentials
        are provided
        """
        backend = LinkedEmailBackend()

        # Can't log in without username or email parameter
        user = backend.authenticate(None)
        self.assertEqual(user, None)
        user = backend.authenticate(self.factory.get("/accounts/login/"))
        self.assertEqual(user, None)

        # Can't log in with incorrect username / email
        user = backend.authenticate(None, username="dne",
                password="johnpass")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="dne@gmail.com",
                password="marypass")
        self.assertEqual(user, None)

        # Can't log in with no password
        user = backend.authenticate(None, username="cas-inst-netid123")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com")
        self.assertEqual(user, None)

        # Can't log in with incorrect password
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password=None)
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password="")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password="marypass")
        self.assertEqual(user, None)

        # Can't log in with username
        user = backend.authenticate(None, email="marysue@outlook.com",
                password="marypass")
        self.assertEqual(user, None)
        user = backend.authenticate(None, username="permid555",
                password="permpass")
        self.assertEqual(user, None)

        # Can't log in with unverified email
        user = backend.authenticate(None, email="pending@gmail.com",
                password="marypass")
        self.assertEqual(user, None)

        # Can't log in as a temporary user
        user = backend.authenticate(None, username="tmp-0123_456",
                password="tmppass")
        self.assertEqual(user, None)


@override_settings(UNIAUTH_ALLOW_SHARED_EMAILS=True)
class UsernameOrLinkedEmailBackendTests(EmailBackendTests):
    """
    Tests the UsernameOrLinkedEmailBackend in backends.py
    """

    def test_username_orlinked_email_backend_valid_credentials(self):
        """
        Ensure backend returns appropriate user when valid
        credentials are provided
        """
        backend = UsernameOrLinkedEmailBackend()

        # Log in with email + password
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password="johnpass")
        self.assertEqual(user, self.john)

        # Log in with email (under "username" parameter) + password
        user = backend.authenticate(None, username="mary.sue@gmail.com",
                password="marypass")
        self.assertEqual(user, self.mary)

        # Log in with verified linked email
        user = backend.authenticate(None, email="alternate@gmail.com",
                password="marypass")
        self.assertEqual(user, self.mary)

        # Log in with shared email address
        user = backend.authenticate(None, email="sharedjones@gmail.com",
                password="mspass")
        self.assertEqual(user, self.ms)
        user = backend.authenticate(None, email="sharedjones@gmail.com",
                password="mrpass")
        self.assertEqual(user, self.mr)

        # Log in with username
        user = backend.authenticate(None, email="marysue@outlook.com",
                password="marypass")
        self.assertEqual(user, self.mary)
        user = backend.authenticate(None, username="permid555",
                password="permpass")
        self.assertEqual(user, self.perm)

    def test_username_or_linked_email_backend_invalid_credentials(self):
        """
        Ensure authentication fails when invalid credentials
        are provided
        """
        backend = UsernameOrLinkedEmailBackend()

        # Can't log in without username or email parameter
        user = backend.authenticate(None)
        self.assertEqual(user, None)
        user = backend.authenticate(self.factory.get("/accounts/login/"))
        self.assertEqual(user, None)

        # Can't log in with incorrect username / email
        user = backend.authenticate(None, username="dne",
                password="johnpass")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="dne@gmail.com",
                password="marypass")
        self.assertEqual(user, None)

        # Can't log in with no password
        user = backend.authenticate(None, username="cas-inst-netid123")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com")
        self.assertEqual(user, None)

        # Can't log in with incorrect password
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password=None)
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password="")
        self.assertEqual(user, None)
        user = backend.authenticate(None, email="johndoe@gmail.com",
                password="marypass")
        self.assertEqual(user, None)

        # Can't log in with unverified email
        user = backend.authenticate(None, email="pending@gmail.com",
                password="marypass")
        self.assertEqual(user, None)

        # Can't log in as a temporary user
        user = backend.authenticate(None, username="tmp-0123_456",
                password="tmppass")
        self.assertEqual(user, None)



try:
    from django.conf.urls import url
except ImportError:
    from django.urls import re_path as url
from uniauth import views

app_name = 'uniauth'

urlpatterns = [
    url(r'^login/$', views.login, name='login'),
    url(r'^cas-login/(?P<institution>[a-z0-9\-]+)/$', views.cas_login, name='cas-login'),
    url(r'^logout/$', views.logout, name='logout'),
    url(r'^signup/$', views.signup, name='signup'),
    url(r'^settings/$', views.settings, name='settings'),
    url(r'^link-to-profile/$', views.link_to_profile, name='link-to-profile'),
    url(r'^link-from-profile/(?P<institution>[a-z0-9\-]+)/$', views.link_from_profile, name='link-from-profile'),
    url(r'^verify-token/(?P<pk_base64>[0-9A-Za-z_\-]+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,32})/$', views.verify_token, name='verify-token'),
    url(r'^password-reset/', views.PasswordReset.as_view(), name='password-reset'),
    url(r'^password-reset-done/', views.PasswordResetDone.as_view(), name='password-reset-done'),
    url(r'^password-reset-verify/(?P<uidb64>[0-9A-Za-z_\-]+)/(?P<token>[0-9A-Za-z]{1,13}-[0-9A-Za-z]{1,32})/', views.PasswordResetVerify.as_view(), name='password-reset-verify'),
    url(r'^password-reset-verify-done/', views.PasswordResetVerifyDone.as_view(), name='password-reset-verify-done'),
    url(r'^jwt-tokens/', views.get_jwt_tokens_from_session, name="jwt-tokens")
]



"""
Django settings for testing.
"""

import os

# Uniauth requires an actual email configuration to be set
# up (to send emails for email validation, changing passwords,
# etc.). This backend just stores emails in django.core.mail.outbox
EMAIL_BACKEND = 'django.core.mail.backends.locmem.EmailBackend'

# Definitely change these for real applications
TESTING = True
SECRET_KEY = 'FAKE_SECRET'
ALLOWED_HOSTS = ['*']

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'uniauth',
    'tests',
]

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

# Use the Uniauth authentication backends
AUTHENTICATION_BACKENDS = [
    'uniauth.backends.LinkedEmailBackend',
    'uniauth.backends.CASBackend',
]

ROOT_URLCONF = 'tests.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [os.path.join(BASE_DIR, 'demo_app', 'templates')],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Database
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

# Password validation
AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

# Internationalization
LANGUAGE_CODE = 'en-us'
TIME_ZONE = 'UTC'
USE_I18N = True
USE_L10N = True
USE_TZ = True

# Static files
STATIC_URL = '/static/'



from cas import CASClient
from django.contrib.auth import authenticate, login as auth_login, \
        logout as auth_logout, REDIRECT_FIELD_NAME
from django.contrib.auth import get_user_model, update_session_auth_hash
from django.contrib.auth.views import PasswordResetCompleteView, \
        PasswordResetConfirmView, PasswordResetDoneView, PasswordResetView
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import EmailMessage
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import Http404, HttpResponseBadRequest, \
        HttpResponseRedirect, JsonResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.urls import reverse, reverse_lazy
from django.urls.exceptions import NoReverseMatch
from django.utils.decorators import method_decorator
from django.views.decorators.cache import never_cache
from django.views.decorators.debug import sensitive_post_parameters
from rest_framework import status
from uniauth.decorators import login_required
from uniauth.forms import AddLinkedEmailForm, ChangePrimaryEmailForm, \
        LinkedEmailActionForm, LoginForm, PasswordChangeForm, \
        PasswordResetForm, SetPasswordForm, SignupForm
from uniauth.merge import merge_model_instances
from uniauth.models import Institution, InstitutionAccount, LinkedEmail
from uniauth.tokens import token_generator, get_jwt_tokens_for_user
from uniauth.utils import choose_username, decode_pk, encode_pk, \
        get_account_username_split, get_protocol, get_random_username, \
        get_redirect_url, get_service_url, get_setting, is_tmp_user, \
        is_unlinked_account
try:
    from urllib import urlencode
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlencode, urlunparse


def _get_global_context(request):
    """
    Returns a dictionary of context data shared by the views
    """
    context = {}

    # Returns the reverse lookup URL of the provided view with
    # the provided slug argument. Returns None if the view is
    # not accessible under the current configuration
    def get_reversed_url_or_none(view_name, slug):
        try:
            return reverse('uniauth:%s' % view_name, args=[slug])
        except NoReverseMatch:
            return None

    # Add a list of Institution tuples, with each containing:
    # (name, slug, CAS login url, Profile link URL)
    institutions = Institution.objects.all()
    institutions = list(map(lambda x: (x.name, x.slug,
            get_reversed_url_or_none("cas-login", x.slug),
            get_reversed_url_or_none("link-from-profile", x.slug)),
            institutions))
    context['institutions'] = institutions

    # Add the query parameters, as a string
    query_params = urlencode(request.GET)
    prefix = '?' if query_params else ''
    context['query_params'] = prefix + query_params

    return context


def _login_success(request, user, next_url, drop_params=[]):
    """
    Determines where to go upon successful authentication:
    Redirects to link tmp account page if user is a temporary
    user, redirects to next_url otherwise.

    Any query parameters whose key exists in drop_params are
    not propogated to the destination URL
    """
    query_params = request.GET.copy()
    jwt_auth = get_setting('UNIAUTH_USE_JWT_AUTH')

    # Drop all blacklisted query parameters
    for key in drop_params:
        if key in query_params:
            del query_params[key]

    # Temporary users should redirect to the link-to-profile page
    if is_tmp_user(user):
        query_params[REDIRECT_FIELD_NAME] = next_url
        return HttpResponseRedirect(reverse('uniauth:link-to-profile') + \
                '?' + urlencode(query_params))

    # All other users should redirect to next_url
    else:
        suffix = ''
        if REDIRECT_FIELD_NAME in query_params:
            del query_params[REDIRECT_FIELD_NAME]
        if len(query_params) > 0:
            suffix = '?' + urlencode(query_params)
        if jwt_auth:
            refresh, access = get_jwt_tokens_for_user(user)
            request.session['jwt-refresh'] = refresh
            request.session['jwt-access'] = access
        return HttpResponseRedirect(next_url + suffix)


def login(request):
    """
    Authenticates the user, then redirects them to the
    next page, defaulting to the URL specified by the
    UNIAUTH_LOGIN_REDIRECT_URL setting.

    Offers users the choice between logging in with their
    Uniauth credentials, or through the CAS interface for
    any supported institution.
    """
    next_url = request.GET.get('next')
    context = _get_global_context(request)

    if not next_url:
        next_url = get_redirect_url(request)

    # If the user is already authenticated, proceed to next page
    if request.user.is_authenticated:
        return _login_success(request, request.user, next_url)

    display_standard = get_setting('UNIAUTH_LOGIN_DISPLAY_STANDARD')
    display_cas = get_setting('UNIAUTH_LOGIN_DISPLAY_CAS')
    num_institutions = len(context['institutions'])

    # Ensure the login settings are configured correctly
    if not display_standard and not display_cas:
        err_msg = "At least one of '%s' and '%s' must be True." % \
                ('UNIAUTH_LOGIN_DISPLAY_STANDARD', 'UNIAUTH_LOGIN_DISPLAY_CAS')
        raise ImproperlyConfigured(err_msg)
    if display_cas and num_institutions == 0:
        err_msg = ("'%s' is True, but there are no Institutions in the "
                "database to log into!") % 'UNIAUTH_LOGIN_DISPLAY_CAS'
        raise ImproperlyConfigured(err_msg)

    context['display_standard'] = display_standard
    context['display_cas'] = display_cas
    context['num_institutions'] = num_institutions

    # If we aren't displaying the standard login form,
    # we're just displaying the CAS login options
    if not display_standard:
        institutions = context['institutions']
        query_params = context['query_params']

        # If there's only one possible institution to log
        # into, redirect to its CAS login page immediately
        if num_institutions == 1:
            return HttpResponseRedirect(institutions[0][2] + query_params)

        # Otherwise, render the page (without the Login form)
        else:
            return render(request, 'uniauth/login.html', context)

    # If we are displaying the login form, and it's
    # a POST request, attempt to validate the form
    elif request.method == "POST":
        form = LoginForm(request, request.POST)

        # Authentication successful: setup session + proceed
        if form.is_valid():
            user = form.get_user()
            auth_login(request, user)
            request.session['auth-method'] = "uniauth"
            return _login_success(request, user, next_url)

        # Authentication failed: render form errors
        else:
            context['form'] = form
            return render(request, 'uniauth/login.html', context)

    # Otherwise, render a blank Login form
    else:
        form = LoginForm(request)
        context['form'] = form
        return render(request, 'uniauth/login.html', context)


def cas_login(request, institution):
    """
    Redirects to the CAS login URL, or verifies the
    CAS ticket, if provided.

    Accepts the slug of the institution to log in to.
    """
    next_url = request.GET.get('next')
    ticket = request.GET.get('ticket')

    # Ensure there is an institution with the provided slug
    try:
        institution = Institution.objects.get(slug=institution)
    except Institution.DoesNotExist:
        raise Http404

    if not next_url:
        next_url = get_redirect_url(request, use_referer=True)

    # If the user is already authenticated, proceed to next page
    if request.user.is_authenticated:
        return _login_success(request, request.user, next_url)

    service_url = get_service_url(request, next_url)
    client = CASClient(version=2, service_url=service_url,
            server_url=institution.cas_server_url)

    # If a ticket was provided, attempt to authenticate with it
    if ticket:
        user = authenticate(request=request, institution=institution,
                ticket=ticket, service=service_url)

        # Authentication successful: setup session + proceed
        if user:
            if not request.session.exists(request.session.session_key):
                request.session.create()
            auth_login(request, user)
            request.session['auth-method'] = "cas-" + institution.slug
            return _login_success(request, user, next_url, ["ticket"])

        # Authentication failed: raise permission denied
        else:
            raise PermissionDenied("Verification of CAS ticket failed.")

    # If no ticket was provided, redirect to the
    # login URL for the institution's CAS server
    else:
        return HttpResponseRedirect(client.get_login_url())


def logout(request):
    """
    Logs the user out of their Uniauth account, and
    redirects to the next page, defaulting to the URL
    specified by the UNIAUTH_LOGOUT_REDIRECT_URL setting.

    If no redirect page is set (URL parameter not given
    and UNIAUTH_LOGOUT_REDIRECT_URL is None), renders the
    logout template.

    Also logs the user out of CAS if they logged in
    via CAS, and the UNIAUTH_LOGOUT_CAS_COMPLETELY
    setting is true.
    """
    next_page = request.GET.get('next')
    auth_method = request.session.get('auth-method')

    if not next_page and get_setting('UNIAUTH_LOGOUT_REDIRECT_URL'):
        next_page = get_redirect_url(request,
                get_setting('UNIAUTH_LOGOUT_REDIRECT_URL'))

    # Formally log out user
    auth_logout(request)

    # Determine whether the user logged in through an institution's CAS
    institution = None
    if auth_method and auth_method.startswith("cas-"):
        try:
            institution = Institution.objects.get(slug=auth_method[4:])
        except Institution.DoesNotExist:
            pass

    # If we need to logout an institution's CAS,
    # redirect to that CAS server's logout URL
    if institution and get_setting('UNIAUTH_LOGOUT_CAS_COMPLETELY'):
        redirect_url = urlunparse(
                (get_protocol(request), request.get_host(),
                    next_page or reverse('uniauth:logout'), '', '', '')
        )
        client = CASClient(version=2, service_url=get_service_url(request),
                server_url=institution.cas_server_url)
        return HttpResponseRedirect(client.get_logout_url(redirect_url))

    # If next page is set, proceed to it
    elif next_page:
        return HttpResponseRedirect(next_page)

    # Otherwise, render the logout view
    else:
        return render(request, 'uniauth/logout.html')


def _send_verification_email(request, to_email, verify_email):
    """
    Sends an email (to to_email) containing a link to
    verify the target email (verify_email) as a linked
    email address for the user.

    Expects to_email as a string, and verify_email as
    a LinkedEmail model instance.
    """
    subject = "Verify your email address."
    message = render_to_string('uniauth/verification-email.html', {
        'protocol': get_protocol(request),
        'domain': get_current_site(request),
        'pk': encode_pk(verify_email.pk),
        'token': token_generator.make_token(verify_email),
        'query_params': _get_global_context(request)["query_params"],
    })
    email = EmailMessage(subject, message, to=[to_email],
            from_email=get_setting('UNIAUTH_FROM_EMAIL'))
    email.send()


def signup(request):
    """
    Creates a new Uniauth profile with the provided
    primary email address and password.

    Prompts user to verify the email address before
    profile is fully created.
    """
    next_url = request.GET.get('next')
    context = _get_global_context(request)

    if not next_url:
        next_url = get_redirect_url(request)

    # If the user is already authenticated + has a Uniauth
    # profile, proceed to next page
    if request.user.is_authenticated and not is_tmp_user(request.user) \
            and not is_unlinked_account(request.user):
        return HttpResponseRedirect(next_url)

    # If it's a POST request, attempt to validate the form
    if request.method == "POST":
        form = SignupForm(request.POST)

        # Validation successful: setup temporary user
        if form.is_valid():
            form_email = form.cleaned_data['email']
            user = request.user

            # If the user is not already authenticated, create a User
            if not user or not user.is_authenticated:
                tmp_username = get_random_username()
                user_model = get_user_model()
                user = user_model.objects.create(username=tmp_username)

            # Set user's password + create linked email
            user.set_password(form.cleaned_data["password1"])
            user.save()
            email, _ = LinkedEmail.objects.get_or_create(
                    profile=user.uniauth_profile, address=form_email,
                    is_verified=False)

            # Send verification email + render waiting template
            _send_verification_email(request, email.address, email)
            return render(request, 'uniauth/verification-waiting.html',
                    {'email': email.address, 'is_signup': True})

        # Validation failed: render form errors
        else:
            context['form'] = form
            return render(request, 'uniauth/signup.html', context)

    # Otherwise, render a blank Signup form
    else:
        form = SignupForm()
        context['form'] = form
        return render(request, 'uniauth/signup.html', context)


@login_required
def settings(request):
    """
    Allows the user to link additional emails, change
    the primary email address of, and link additional
    institution accounts to their Uniauth profile.
    """
    context = _get_global_context(request)
    context['email_resent'] = None
    context['email_added'] = None
    context['password_changed'] = False

    action_email_form = None
    add_email_form = None
    change_email_form = None
    change_password_form = None

    # This page may only be accessed by users with Uniauth profiles:
    # if the user is logged in with an unlinked InsitutionAccount,
    # redirect them to the link page
    if is_unlinked_account(request.user):
        params = urlencode({'next': reverse('uniauth:settings')})
        return HttpResponseRedirect(reverse('uniauth:link-to-profile') + \
                '?' + params)

    # If it's a POST request, determine which form was submitted
    if request.method == 'POST':

        # Linked Email Action Form submitted: validate the
        # delete / resend action + perform it if able
        if request.POST.get('action-email-submitted'):
            action_email_form = LinkedEmailActionForm(request.POST)
            if action_email_form.is_valid():
                delete_pk = action_email_form.cleaned_data.get("delete_pk")
                resend_pk = action_email_form.cleaned_data.get("resend_pk")

                # If they asked to delete a linked email: it must
                # belong to the user and not be the primary address
                if delete_pk:
                    try:
                        email = LinkedEmail.objects.get(pk=delete_pk)
                        if email.profile.user == request.user and \
                                email.address != request.user.email:
                            email.delete()
                    except LinkedEmail.DoesNotExist:
                        pass

                # If they asked to resend a verification email: it
                # must belong to the user and be pending verification
                elif resend_pk:
                    try:
                        email = LinkedEmail.objects.get(pk=resend_pk)
                        if email.profile.user == request.user and \
                                not email.is_verified:
                            _send_verification_email(request,
                                    email.address, email)
                            context['email_resent'] = email.address
                    except LinkedEmail.DoesNotExist:
                        pass

                action_email_form = None

        # Add Linked Email Form submitted: validate and create
        # new linked email + send verification email
        elif request.POST.get('add-email-submitted'):
            add_email_form = AddLinkedEmailForm(request.user, request.POST)
            if add_email_form.is_valid():
                email = LinkedEmail.objects.create(
                        profile=request.user.uniauth_profile,
                        address=add_email_form.cleaned_data["email"],
                        is_verified=False)
                _send_verification_email(request, email.address, email)
                context['email_added'] = email.address
                add_email_form = None

        # Change Primary Email Address Form submitted:
        # validate and set user's primary email address
        elif request.POST.get('change-email-submitted'):
            change_email_form = ChangePrimaryEmailForm(request.user,
                    request.POST)
            if change_email_form.is_valid():
                request.user.email = change_email_form.cleaned_data["email"]
                request.user.save()
                change_email_form = None

        # Change Password Form submitted: validate and set password
        elif request.POST.get('change-password-submitted'):
            change_password_form = PasswordChangeForm(request.user,
                    request.POST)
            if change_password_form.is_valid():
                change_password_form.save()
                update_session_auth_hash(request, request.user)
                context['password_changed'] = True
                change_password_form = None

    # Create blank versions of all non-bound forms
    if action_email_form is None:
        action_email_form = LinkedEmailActionForm(request.user)
    if add_email_form is None:
        add_email_form = AddLinkedEmailForm(request.user)
    if change_email_form is None:
        change_email_form = ChangePrimaryEmailForm(request.user)
    if change_password_form is None:
        change_password_form = PasswordChangeForm(request.user)

    # Assemble the context and render
    context['action_email_form'] = action_email_form
    context['add_email_form'] = add_email_form
    context['change_email_form'] = change_email_form
    context['change_password_form'] = change_password_form
    return render(request, 'uniauth/settings.html', context)


def _add_institution_account(profile, slug, cas_id):
    """
    Accepts an institution slug and cas ID and links an
    InsitutionAccount to the provided Uniauth user.
    """
    institution = Institution.objects.get(slug=slug)
    InstitutionAccount.objects.create(profile=profile,
            institution=institution, cas_id=cas_id)


def link_to_profile(request):
    """
    If the user is a temporary one who was logged in via
    an institution (not through a Uniauth profile), offers
    them the choice between logging to an existing Uniauth
    account or creating a new one.

    The institution account is (eventually) linked to the
    Uniauth profile the user logged into / created.
    """
    next_url = request.GET.get('next')
    context = _get_global_context(request)

    if not next_url:
        next_url = get_redirect_url(request)

    params = urlencode({'next': next_url})
    context['next_url'] = next_url

    # If the user is not authenticated at all, redirect to login page
    if not request.user.is_authenticated:
        return HttpResponseRedirect(reverse('uniauth:login') + '?' + params)

    # If the user is already authenticated + verified, proceed to next page
    if not is_tmp_user(request.user) and not is_unlinked_account(request.user):
        return HttpResponseRedirect(next_url)

    # If the user is temporary, but was not logged in via an institution
    # (e.g. created through Uniauth, but not verified), redirect to signup
    if not is_unlinked_account(request.user):
        return HttpResponseRedirect(reverse('uniauth:signup') + '?' + params)

    # At this point, we've ensured the user is temporary and was
    # logged in via an institution. We just need to handle the
    # Login Form, if the user chooses to link to an existing account.

    # If it's a POST request, attempt to validate the form
    if request.method == "POST":
        form = LoginForm(request, request.POST)

        # Authentication successful
        if form.is_valid():
            unlinked_user = request.user
            username_split = get_account_username_split(request.user.username)

            # Log in as the authenticated Uniauth user
            user = form.get_user()
            auth_login(request, user)

            # Merge the unlinked account into the logged in profile,
            # then add the institution account described by the username
            merge_model_instances(user, [unlinked_user])
            _add_institution_account(user.uniauth_profile, username_split[1],
                    username_split[2])

            slug = username_split[1]
            context['institution'] = Institution.objects.get(slug=slug)
            return render(request, 'uniauth/link-success.html', context)

        # Authentication failed: render form errors
        else:
            context['form'] = form
            return render(request, 'uniauth/link-to-profile.html', context)

    # Otherwise, render a blank Login form
    else:
        form = LoginForm(request)
        context['form'] = form
        return render(request, 'uniauth/link-to-profile.html', context)


def link_from_profile(request, institution):
    """
    Attempts to authenticate a CAS account for the provided
    institution, and links it to the current Uniauth profile
    if successful.
    """
    next_url = request.GET.get('next')
    ticket = request.GET.get('ticket')

    # Ensure there is an institution with the provided slug
    try:
        institution = Institution.objects.get(slug=institution)
    except Institution.DoesNotExist:
        raise Http404

    if not next_url:
        next_url = get_redirect_url(request, use_referer=True)

    # If the user is not already logged into a verified
    # Uniauth account, raise permission denied
    if not request.user.is_authenticated or is_tmp_user(request.user) \
            or is_unlinked_account(request.user):
        raise PermissionDenied("Must be logged in as verified Uniauth user.")

    service_url = get_service_url(request, next_url)
    client = CASClient(version=2, service_url=service_url,
            server_url=institution.cas_server_url)

    # If a ticket was provided, attempt to authenticate with it
    if ticket:
        user = authenticate(request=request, institution=institution,
                ticket=ticket, service=service_url)

        # Authentication successful: link to Uniauth profile if
        # the institution account has not been linked yet + proceed
        if user:
            if is_unlinked_account(user):
                merge_model_instances(request.user, [user])
                username_split = get_account_username_split(user.username)
                _add_institution_account(request.user.uniauth_profile,
                        username_split[1], username_split[2])

            return HttpResponseRedirect(next_url)

        # Authentication failed: raise permission denied
        else:
            raise PermissionDenied("Verification of CAS ticket failed")

    # If no ticket was provided, redirect to the
    # login URL for the institution's CAS server
    else:
        return HttpResponseRedirect(client.get_login_url())


def verify_token(request, pk_base64, token):
    """
    Verifies a token generated for validating an email
    address, and notifies the user whether verification
    was successful.
    """
    next_url = request.GET.get('next') or request.GET.get(REDIRECT_FIELD_NAME)
    context = {'next_url': next_url, 'is_signup': False}
    user_model = get_user_model()

    # Attempt to get the linked email to verify
    try:
        email_pk = decode_pk(pk_base64)
        email = LinkedEmail.objects.get(pk=email_pk)
    except (TypeError, ValueError, OverflowError, user_model.DoesNotExist,
            LinkedEmail.DoesNotExist):
        email = None

    # In the unlikely scenario that a user is trying to sign up
    # with an email another verified user has as a primary email
    # address, reject verification immediately
    if email is not None and is_tmp_user(email.profile.user) and \
            get_user_model().objects.filter(email=email.address).exists():
        email = None

    # If the token successfully verifies, update the linked email
    if email is not None and token_generator.check_token(email, token):
        email.is_verified = True
        email.save()

        # If the user this email is linked to is a temporary
        # one, change it to a fully registered user
        user = email.profile.user
        if is_tmp_user(user) or is_unlinked_account(user):
            context['is_signup'] = True
            old_username = user.username

            # Change the email + username to the verified email
            user.email = email.address
            user.username = choose_username(user.email)
            user.save()

            # If the user was created via CAS, add the institution
            # account described by the temporary username
            if old_username.startswith("cas"):
                username_split = get_account_username_split(old_username)
                _add_institution_account(user.uniauth_profile,
                        username_split[1], username_split[2])

        # If UNIAUTH_ALLOW_SHARED_EMAILS is False, and there were
        # pending LinkedEmails for this address on other accounts,
        # delete them
        if not get_setting('UNIAUTH_ALLOW_SHARED_EMAILS'):
            LinkedEmail.objects.filter(address=email.address,
                    is_verified=False).delete()

        return render(request, 'uniauth/verification-success.html', context)

    # If anything went wrong, just render the failed verification template
    else:
        return render(request, 'uniauth/verification-failure.html', context)


# The password reset views are pulled from the django.conrib.auth
# package, and are used largely unmodified. We just set things like
# the template and reverse URL names to use Uniauth's files + naming
# scheme.

class PasswordReset(PasswordResetView):
    """
    Asks user for an email address to send the link to, then
    sends a password reset link to that address.

    If no user has that email address linked to their account,
    no email is sent, unbeknownst to the user.
    """
    email_template_name = 'uniauth/password-reset-email.html'
    form_class = PasswordResetForm
    from_email = get_setting('UNIAUTH_FROM_EMAIL')
    success_url = reverse_lazy('uniauth:password-reset-done')
    template_name = 'uniauth/password-reset.html'

    def form_valid(self, form):
        """
        Add global context data before composing reset email
        """
        self.extra_email_context = _get_global_context(self.request)
        return super(PasswordReset, self).form_valid(form)


class PasswordResetDone(PasswordResetDoneView):
    """
    Shows success message once email was sent.
    """
    template_name = 'uniauth/password-reset-done.html'


class PasswordResetVerify(PasswordResetConfirmView):
    """
    Verifies the provided token, and prompts user for their
    new password if successful.
    """
    form_class = SetPasswordForm
    success_url = reverse_lazy('uniauth:password-reset-verify-done')
    template_name = 'uniauth/password-reset-verify.html'

    @method_decorator(sensitive_post_parameters())
    @method_decorator(never_cache)
    def dispatch(self, *args, **kwargs):
        """
        Save query params in session if available.
        """
        query_params = _get_global_context(self.request)['query_params']
        if query_params:
            self.request.session['password-reset-query-params'] = query_params
        return super(PasswordResetVerify, self).dispatch(*args, **kwargs)

    def form_valid(self, form):
        """
        Set success_url to include query params saved in session
        """
        query_params = self.request.session\
                .pop('password-reset-query-params', '')
        self.success_url += query_params
        return super(PasswordResetVerify, self).form_valid(form)


class PasswordResetVerifyDone(PasswordResetCompleteView):
    """
    Shows success message once new password is set.
    """
    template_name = 'uniauth/password-reset-verify-done.html'

    def get_context_data(self, **kwargs):
        """
        Add global context data to template context.
        """
        context = super(PasswordResetVerifyDone, self)\
                .get_context_data(**kwargs)
        context.update(_get_global_context(self.request))
        return context


def get_jwt_tokens_from_session(request):
    if request.method == "GET":
        refresh = request.session.pop("jwt-refresh", None)
        access = request.session.pop("jwt-access", None)
        if refresh is None or access is None:
            return JsonResponse({}, status=status.HTTP_404_NOT_FOUND)
        else:
            return JsonResponse({
                "refresh": refresh,
                "access": access
            }, status=status.HTTP_200_OK)



from datetime import timedelta
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_save
from django.dispatch import receiver
from django.utils import timezone


class UserProfile(models.Model):
    """
    Extension of the default User model containing
    extra information necessary for UniAuth to run.
    """

    # User this profile is extending
    user = models.OneToOneField(get_user_model(),
            related_name='uniauth_profile',
            on_delete=models.CASCADE, null=False)

    def get_display_id(self):
        """
        Returns a display-friendly ID for this User, using their
        username. Users created via CAS authentication will have
        their CAS ID returned (without the cas-institution prefix),
        and Users with an email address for a username will have
        the string preceeding the "@" returned. All other users
        will have their raw username returned.

        Note that these IDs are not guaranteed to be unique.
        """
        username = self.user.username
        if "@" in username:
            return username.split("@")[0]
        if username.startswith('cas-'):
            from uniauth.utils import get_account_username_split
            return get_account_username_split(username)[-1]
        return username

    def __str__(self):
        try:
            return self.user.email or self.user.username
        except:
            return "NULL"


@receiver(post_save, sender=get_user_model())
def create_user_profile(sender, instance, created, **kwargs):
    """
    Create a Uniauth profile automatically when a User is created.

    If the User was given an email on creation, add it as a verified
    LinkedEmail immediately.
    """
    if created:
        profile = UserProfile.objects.create(user=instance)
        if profile and instance.email:
            LinkedEmail.objects.create(profile=profile,
                    address=instance.email, is_verified=True)


@receiver(post_save, sender=get_user_model())
def clear_old_tmp_users(sender, instance, created, **kwargs):
    """
    Deletes temporary users more than PASSWORD_RESET_TIMEOUT_DAYS
    old when a User is created.

    Does nothing if the user model does not have date_joined field.
    """
    if created:
        user_model = get_user_model()
        if hasattr(user_model, 'date_joined'):
            from uniauth.utils import get_setting
            timeout_days = timedelta(
                    days=get_setting("PASSWORD_RESET_TIMEOUT_DAYS"))
            tmp_expire_date = (timezone.now() - timeout_days).replace(
                    hour=0, minute=0, second=0, microsecond=0)
            user_model.objects.filter(username__startswith='tmp-',
                    date_joined__lt=tmp_expire_date).delete()


class LinkedEmail(models.Model):
    """
    Represents an email address linked to a user's account.
    """

    # Person owning this email
    profile = models.ForeignKey('UserProfile', related_name='linked_emails',
            on_delete=models.CASCADE, null=False)

    # The email address
    address = models.EmailField(null=False, blank=False)

    # Whether the linked email is verified
    is_verified = models.BooleanField(default=False)

    def clean(self):
        """
        Ensures an email can't be linked and verified for multiple
        accounts if UNIAUTH_ALLOW_SHARED_EMAILS is False.

        Also ensures a user does not link more than the maximum
        number of linked emails per user.
        """
        from uniauth.utils import get_setting

        # Check for shared emails if necessary
        if not get_setting('UNIAUTH_ALLOW_SHARED_EMAILS') and \
                LinkedEmail.objects.filter(address=self.address,
                        is_verified=True) and self.is_verified:
            raise ValidationError("This email address has already been " +
                    "linked to another account.")

        # Ensure a user doesn't link more than the maximum
        max_linked_emails = get_setting('UNIAUTH_MAX_LINKED_EMAILS')
        if max_linked_emails > 0 and \
                self.profile.linked_emails.count() >= max_linked_emails:
            raise ValidationError(("You can not link more than %d emails "
                "to your account.") % max_linked_emails)
        super(LinkedEmail, self).clean()

    def __str__(self):
        try:
            return "%s | %s" % (self.profile, self.address)
        except:
            return "NULL"


class Institution(models.Model):
    """
    Represents an organization holding a CAS server
    that can be logged into.
    """

    # Name of the institution
    name = models.CharField(max_length=30, null=False, blank=False)

    # Slugified version of name
    slug = models.CharField(max_length=30, null=False, blank=False,
            unique=True)

    # CAS server location
    cas_server_url = models.URLField(null=False, blank=False)

    def __str__(self):
        try:
            return self.slug
        except:
            return "NULL"


class InstitutionAccount(models.Model):
    """
    Relates users to the accounts they have at
    institutions, and stores any associated data.
    """

    # The person the account is for
    profile = models.ForeignKey('UserProfile', related_name='accounts',
            on_delete=models.CASCADE, null=False)

    # The institution the account is with
    institution = models.ForeignKey('Institution', related_name='accounts',
            on_delete=models.CASCADE, null=False)

    # The ID used by the CAS server
    cas_id = models.CharField(max_length=30, null=False, blank=False)

    class Meta:
        unique_together = ('institution', 'cas_id')

    def __str__(self):
        try:
            return "%s | %s | account" % (self.profile, self.institution)
        except:
            return "NULL"



from django.contrib.auth.models import User
from django.test import override_settings, TestCase
from uniauth.models import Institution, InstitutionAccount, LinkedEmail
from uniauth.merge import merge_model_instances


class MergeModelInstancesTests(TestCase):
    """
    Tests the merge_model_instances method in merge.py
    """

    def setUp(self):
        self.inst1 = Institution.objects.create(name="Test Uni",
                slug="test-uni", cas_server_url="https://cas.testuni.edu")
        self.inst2 = Institution.objects.create(name="Other Inst",
                slug="other-inst", cas_server_url="https://fed.other-inst.edu")

        self.user1 = User.objects.create(username="user1@example.com",
                email="user1@example.com")
        self.user2 = User.objects.create(username="cas-test-uni-user2")
        InstitutionAccount.objects.create(profile=self.user2.uniauth_profile,
                institution=self.inst1, cas_id="user2")

        self.user3 = User.objects.create(username="user3@gmail.com",
                email="user3@gmail.com")
        LinkedEmail.objects.create(profile=self.user3.uniauth_profile,
                address="backup@gmail.com")
        InstitutionAccount.objects.create(profile=self.user3.uniauth_profile,
                institution=self.inst1, cas_id="john123")
        self.user4 = User.objects.create(username="us.er4@nowhere.gov",
                email="huh@what.edu")
        LinkedEmail.objects.create(profile=self.user4.uniauth_profile,
                address="another@gmail.com", is_verified=True)
        InstitutionAccount.objects.create(profile=self.user4.uniauth_profile,
                institution=self.inst2, cas_id="exid001")
        self.user5 = User.objects.create(username="user5@foo.bar",
                email="user5@foo.bar")

    def _check_emails(self, actual, expected):
        actual_values = [(x.profile, x.address) for x in actual]
        expected_values = [(x.profile, x.address) for x in expected]
        self.assertEqual(actual_values, expected_values)

    def _check_accounts(self, actual, expected):
        act_values = [(x.profile, x.institution, x.cas_id) for x in actual]
        exp_values = [(x.profile, x.institution, x.cas_id) for x in expected]
        self.assertEqual(act_values, exp_values)

    @override_settings(UNIAUTH_PERFORM_RECURSIVE_MERGING=False)
    def test_merge_model_instances_non_recursive(self):
        """
        Ensure non-recursive merges work as expected
        """
        merge_model_instances(self.user1, [self.user2])
        self.assertTrue(User.objects.filter(username="user1@example.com")\
                .exists())
        self.assertFalse(User.objects.filter(username="cas-test-uni-user2")\
                .exists())

        merged = User.objects.get(username="user1@example.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
                LinkedEmail(profile=merged.uniauth_profile,
                        address="user1@example.com"),
        ]
        expected_accounts = [
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)

        merge_model_instances(self.user3, [self.user4, self.user5])
        self.assertTrue(User.objects.filter(username="user3@gmail.com")\
                .exists())
        self.assertFalse(User.objects.filter(username="us.er4@nowhere.gov")\
                .exists())
        self.assertFalse(User.objects.filter(username="user5@foo.bar")\
                .exists())

        merged = User.objects.get(username="user3@gmail.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
                LinkedEmail(profile=merged.uniauth_profile,
                        address="backup@gmail.com"),
                LinkedEmail(profile=merged.uniauth_profile,
                    address="user3@gmail.com"),
        ]
        expected_accounts = [
                InstitutionAccount(profile=merged.uniauth_profile,
                        institution=self.inst1, cas_id="john123"),
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)

    @override_settings(UNIAUTH_PERFORM_RECURSIVE_MERGING=True)
    def test_merge_model_instances_recursive(self):
        """
        Ensure recursive merges work as expected
        """
        merge_model_instances(self.user1, [self.user2])
        self.assertTrue(User.objects.filter(username="user1@example.com")\
                .exists())
        self.assertFalse(User.objects.filter(username="cas-test-uni-user2")\
                .exists())

        merged = User.objects.get(username="user1@example.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
                LinkedEmail(profile=merged.uniauth_profile,
                        address="user1@example.com"),
        ]
        expected_accounts = [
                InstitutionAccount(profile=merged.uniauth_profile,
                        institution=self.inst1, cas_id="user2"),
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)

        merge_model_instances(self.user3, [self.user4, self.user5])
        self.assertTrue(User.objects.filter(username="user3@gmail.com")\
                .exists())
        self.assertFalse(User.objects.filter(username="us.er4@nowhere.gov")\
                .exists())
        self.assertFalse(User.objects.filter(username="user5@foo.bar")\
                .exists())

        merged = User.objects.get(username="user3@gmail.com")
        emails = list(merged.uniauth_profile.linked_emails.order_by("address"))
        accounts = list(merged.uniauth_profile.accounts.order_by("cas_id"))

        expected_emails = [
                LinkedEmail(profile=merged.uniauth_profile,
                        address="another@gmail.com"),
                LinkedEmail(profile=merged.uniauth_profile,
                        address="backup@gmail.com"),
                LinkedEmail(profile=merged.uniauth_profile,
                        address="huh@what.edu"),
                LinkedEmail(profile=merged.uniauth_profile,
                        address="user3@gmail.com"),
                LinkedEmail(profile=merged.uniauth_profile,
                        address="user5@foo.bar"),
        ]
        expected_accounts = [
                InstitutionAccount(profile=merged.uniauth_profile,
                        institution=self.inst2, cas_id="exid001"),
                InstitutionAccount(profile=merged.uniauth_profile,
                        institution=self.inst1, cas_id="john123"),
        ]
        self._check_emails(emails, expected_emails)
        self._check_accounts(accounts, expected_accounts)




from django.contrib import admin
from uniauth import models

admin.site.register(models.UserProfile)
admin.site.register(models.LinkedEmail)
admin.site.register(models.Institution)
admin.site.register(models.InstitutionAccount)



from django.contrib.auth.models import User
from django.core import mail
from django.http import HttpResponseRedirect
from django.test import override_settings, RequestFactory, TestCase
from django.urls import reverse
import json
from tests.utils import assert_urls_equivalent
from uniauth.models import Institution, InstitutionAccount, LinkedEmail
from uniauth.views import _add_institution_account, _get_global_context, \
        _login_success, _send_verification_email, get_jwt_tokens_from_session


class AddInstitutionAccountTests(TestCase):
    """
    Tests the _add_instituion_account method in views.py
    """

    def setUp(self):
        self.inst1 = Institution.objects.create(name="Test Uni",
                slug="test-uni", cas_server_url="https://cas.testuni.edu")
        self.inst2 = Institution.objects.create(name="Other Inst",
                slug="other-inst", cas_server_url="https://fed.other-inst.edu")

    def test_add_institution_account_link_succeeds(self):
        """
        Ensures the method successfully links the new
        Institution Account to the profile
        """
        user = User.objects.create(username="johndoe")
        User.objects.create(username="cas-test-uni-jd123")
        _add_institution_account(user.uniauth_profile, "test-uni", "jd123")
        linked_accounts = user.uniauth_profile.accounts.all()
        cas_user1 = User.objects.get(username="cas-test-uni-jd123")
        self.assertEqual(len(linked_accounts), 1)
        self.assertEqual(linked_accounts[0].institution, self.inst1)
        self.assertEqual(linked_accounts[0].cas_id, "jd123")

        User.objects.create(username="cas-other-inst-john.doe4")
        _add_institution_account(user.uniauth_profile, "other-inst",
                "john.doe4")
        linked_accounts = user.uniauth_profile.accounts.all()
        cas_user2 = User.objects.get(username="cas-other-inst-john.doe4")
        self.assertEqual(len(linked_accounts), 2)
        self.assertEqual(linked_accounts[1].institution, self.inst2)
        self.assertEqual(linked_accounts[1].cas_id, "john.doe4")


class GetGlobalContextTests(TestCase):
    """
    Tests the _get_global_context method in views.py
    """

    def setUp(self):
        self.factory = RequestFactory()
        Institution.objects.all().delete()
        self.inst1 = Institution.objects.create(name="Test Uni",
                slug="test-uni", cas_server_url="https://cas.testuni.edu")
        self.inst2 = Institution.objects.create(name="Other Inst",
                slug="other-inst", cas_server_url="https://fed.other-inst.edu")

    def test_get_global_context_correct(self):
        """
        Ensure method returns the expected template context
        """
        request = self.factory.get("/accounts/login/",
                data={"foo": "bar", "next": "/next-page/"})
        result = _get_global_context(request)
        result["institutions"].sort(key=lambda x: x[0])
        expected_institutions = [
                ("Other Inst", "other-inst", "/accounts/cas-login/other-inst/",
                        "/accounts/link-from-profile/other-inst/"),
                ("Test Uni", "test-uni", "/accounts/cas-login/test-uni/",
                        "/accounts/link-from-profile/test-uni/"),
        ]
        expected_params = "?foo=bar&next=%2Fnext-page%2F"
        self.assertEqual(sorted(list(result.keys())),
                ["institutions", "query_params"])
        self.assertEqual(result["institutions"], expected_institutions)
        assert_urls_equivalent(result["query_params"],
                expected_params, self.assertEqual)


class LoginSuccessTests(TestCase):
    """
    Tests the _login_success method in views.py
    """

    factory = RequestFactory()

    def _run_test(self, username, params_dict, expected_url):
        user = User.objects.create(username=username)
        request = self.factory.get("/test/page/", data=params_dict)
        result = _login_success(request, user, "/next-page/")
        self.assertEqual(type(result), HttpResponseRedirect)
        assert_urls_equivalent(result.url, expected_url, self.assertEqual)

    def test_login_success_correct_destination(self):
        """
        Ensure method returns the correct destination URL,
        dependent on whether the user has a temporary account
        """
        self._run_test("tmp-abc123", {},
                "/accounts/link-to-profile/?next=%2Fnext-page%2F")
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=False):
            self._run_test("cas-dashed-inst-id123", {},
                    "/accounts/link-to-profile/?next=%2Fnext-page%2F")
        with self.settings(UNIAUTH_ALLOW_STANDALONE_ACCOUNTS=True):
            self._run_test("cas-dashed-inst-id987", {}, "/next-page/")
        self._run_test("john.doe@hmail.moc", {}, "/next-page/")

    def test_login_success_preserve_query_params(self):
        """
        Ensure outputted URL preserves query parameters
        """
        self._run_test("tmp-a0b1c2", {"hello": "world"},
                "/accounts/link-to-profile/?hello=world&next=%2Fnext-page%2F")
        self._run_test("tmp-z9y8x7", {"foo": "bar", "next": "/detour/"},
                "/accounts/link-to-profile/?foo=bar&next=%2Fnext-page%2F")
        self._run_test("example@foo.bar", {"hi": "bye"}, "/next-page/?hi=bye")
        self._run_test("student@inst.edu", {"dead": "beef", "next": "/detour/"},
                "/next-page/?dead=beef")

    @override_settings(UNIAUTH_USE_JWT_AUTH=True)    
    def test_login_success_use_jwt_auth(self):
        user = User.objects.create(username="student@institution.edu")

        self.client.force_login(user)
        session = self.client.session
        
        response = self.client.get(reverse('uniauth:login'))
     
        self.assertEqual(type(response), HttpResponseRedirect)
        self.assertTrue(session['jwt-refresh'])
        self.assertTrue(session['jwt-access'])


class SendVerificationEmail(TestCase):
    """
    Tests the _send_verification_email method in views.py
    """

    factory = RequestFactory()

    @override_settings(UNIAUTH_FROM_EMAIL="uniauth@testsmtp.ml")
    def test_send_verification_email_correct(self):
        """
        Ensure the subject, to address, and from address for
        a sent verification email is correct
        """
        user = User.objects.create(username="student")
        verify_email = LinkedEmail.objects.create(profile=user.uniauth_profile,
                address="toverify@gmail.com", is_verified=False)
        request = self.factory.get("/accounts/somepage/", secure=False)
        _send_verification_email(request, "student@example.edu", verify_email)
        self.assertEqual(len(mail.outbox), 1)
        self.assertTrue("verify" in mail.outbox[0].subject.lower())
        self.assertTrue("/verify-token/" in mail.outbox[0].body)
        self.assertEqual(mail.outbox[0].to, ["student@example.edu"])
        self.assertEqual(mail.outbox[0].from_email, "uniauth@testsmtp.ml")

        user2 = User.objects.create(username="otherstud")
        verify_email2 = LinkedEmail.objects.create(
                profile=user2.uniauth_profile, address="stud_2@example.edu",
                is_verified=False)
        request = self.factory.get("/accounts/otherpage/", secure=True)
        _send_verification_email(request, "stud_2@example.edu", verify_email2)
        self.assertEqual(len(mail.outbox), 2)
        self.assertTrue("verify" in mail.outbox[1].subject.lower())
        self.assertTrue("/verify-token/" in mail.outbox[1].body)
        self.assertEqual(mail.outbox[1].to, ["stud_2@example.edu"])
        self.assertEqual(mail.outbox[1].from_email, "uniauth@testsmtp.ml")


class GetJWTTokensFromSession(TestCase):
    """
    Tests the get_jwt_tokens_from_session method in views.py
    """   

    factory = RequestFactory()
        
    FAKE_REFRESH_TOKEN = "refresh.token.string"
    FAKE_ACCESS_TOKEN = "access.token.string" 


    def _run_test(self, username, session_state, expected_response_status, expected_response_data):
        user = User.objects.create(username=username)
        request = self.factory.get("/jwt-tokens/", data={})
        request.user = user
        request.session = session_state
        response = get_jwt_tokens_from_session(request)
        self.assertEqual(response.status_code, expected_response_status)
        self.assertEqual(json.loads(response.content.decode('utf-8')), expected_response_data)


    def test_get_jwt_tokens_from_session(self):
        self._run_test("tmp-c65433", {"jwt-refresh": self.FAKE_REFRESH_TOKEN, "jwt-access": self.FAKE_ACCESS_TOKEN}, 200, {"refresh": self.FAKE_REFRESH_TOKEN, "access": self.FAKE_ACCESS_TOKEN})
        self._run_test("tmp-a67653", {"jwt-refresh": self.FAKE_REFRESH_TOKEN}, 404, {})
        self._run_test("tmp-a65343", {"jwt-access": self.FAKE_ACCESS_TOKEN}, 404, {})
        self._run_test("tmp-a75643", {}, 404, {})





"""
This command is used to add an Institution to the database.

Execution: python manage.py add_institution <name> <cas_server_url>
"""

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.core.validators import URLValidator
from django.utils.text import slugify
from uniauth.models import Institution

class Command(BaseCommand):
    help = "Adds an institution to the database."

    def add_arguments(self, parser):
        parser.add_argument('name')
        parser.add_argument('cas_server_url')
        parser.add_argument(
            '--update-existing',
            action='store_true',
            default=False,
            help='Update the institution, if it already exists.')

    def handle(self, *args, **options):
        slug = slugify(options['name'])
        cas_server_url = options['cas_server_url']

        if (not options['update_existing']
            and Institution.objects.filter(slug=slug).exists()):
            raise CommandError("An institution with slug '" +
                        slug + "' already exists.")

        try:
            validator = URLValidator()
            validator(options['cas_server_url'])
        except ValidationError:
            raise CommandError("Provided CAS server URL '" +
                    cas_server_url + "' is malformed.")
      
        institution, created = Institution.objects.get_or_create(
            name=options['name'],
            slug=slug,
            defaults={'cas_server_url': cas_server_url}
            )

        if created:
            self.stdout.write("Created institution '%s'.\n" % str(institution))
        elif institution.cas_server_url != cas_server_url:
            # If institution already exists but with a different URL,
            # update it.
            institution.cas_server_url = cas_server_url
            institution.save()
            self.stdout.write("Updated institution '%s'.\n" % str(institution))



"""
URL configuration for testing.
"""

try:
    from django.conf.urls import include, url
except ImportError:
    from django.urls import include, re_path as url

urlpatterns = [
    url('accounts/', include('uniauth.urls', namespace='uniauth')),
]



# Generated by Django 2.2.2 on 2019-07-07 23:05

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('uniauth', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='userprofile',
            name='user',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='uniauth_profile', to=settings.AUTH_USER_MODEL),
        ),
    ]



import setuptools

with open("README.md", "r") as readme:
    long_description = readme.read()

setuptools.setup(
    name="django-uniauth",
    version="1.3.2",
    author="Lance Goodridge",
    author_email="ldgoodridge95@gmail.com",
    keywords=["django", "auth", "authentication", "cas", "sso", "single sign-on"],
    description="A Django app for managing CAS and custom user authentication.",
    include_package_data=True,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lgoodridge/django-uniauth",
    license='LGPLv3',
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*",
    install_requires=[
        "Django>=1.11",
        "python-cas>=1.4.0",
        "djangorestframework-simplejwt>=4.1.0",
        "PyJWT<=1.7.1"
    ],
    extras_require = {
        ":python_version<='3.2'": ["mock"],
    },
    packages=setuptools.find_packages(exclude=["demo-app",]),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 1.11",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ]
)



from datetime import timedelta
from django.conf import settings
from django.contrib.auth import get_user_model, REDIRECT_FIELD_NAME
from django.shortcuts import resolve_url
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.encoding import force_bytes
try:
    from django.utils.encoding import force_text
except ImportError:
    from django.utils.encoding import force_str as force_text
from django.utils.http import urlsafe_base64_decode, urlsafe_base64_encode
try:
    from urllib import urlencode
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlencode, urlunparse


# The default value for all settings used by Uniauth
DEFAULT_SETTING_VALUES = {
    'LOGIN_URL': '/accounts/login/',
    'PASSWORD_RESET_TIMEOUT_DAYS': 3,
    'UNIAUTH_ALLOW_STANDALONE_ACCOUNTS': True,
    'UNIAUTH_ALLOW_SHARED_EMAILS': True,
    'UNIAUTH_FROM_EMAIL': 'uniauth@example.com',
    'UNIAUTH_LOGIN_DISPLAY_STANDARD': True,
    'UNIAUTH_LOGIN_DISPLAY_CAS': True,
    'UNIAUTH_LOGIN_REDIRECT_URL': '/',
    'UNIAUTH_LOGOUT_CAS_COMPLETELY': False,
    'UNIAUTH_LOGOUT_REDIRECT_URL': None,
    'UNIAUTH_MAX_LINKED_EMAILS': 20,
    'UNIAUTH_PERFORM_RECURSIVE_MERGING': True,
    'UNIAUTH_USE_JWT_AUTH': False
}


def choose_username(email):
    """
    Chooses a unique username for the provided user.

    Sets the username to the email parameter umodified if
    possible, otherwise adds a numerical suffix to the email.
    """
    def get_suffix(number):
        return "" if number == 1 else "_"+str(number).zfill(3)
    user_model = get_user_model()
    num = 1
    while user_model.objects.filter(username=email+get_suffix(num)).exists():
        num += 1
    return email + get_suffix(num)


def decode_pk(encoded_pk):
    """
    Decodes the provided base64 encoded pk into its
    original value, as a string
    """
    return force_text(urlsafe_base64_decode(encoded_pk))


def encode_pk(pk):
    """
    Returns the base64 encoding of the provided pk
    """
    encoded = urlsafe_base64_encode(force_bytes(pk))
    # On Django <2.1, this method returns a byte string
    try:
        encoded = encoded.decode()
    except AttributeError:
        pass
    return encoded


def flush_old_tmp_users(days=1):
    """
    Delete temporary users more than the specified number of days old.

    Returns the number of users deleted by this action.
    """
    user_model = get_user_model()
    old_tmp_users = user_model.objects.filter(
        username__startswith="tmp-",
        date_joined__lte=timezone.now()-timedelta(days=days)
    )
    num_deleted = old_tmp_users.count()
    old_tmp_users.delete()
    return num_deleted


def get_account_username_split(username):
    """
    Accepts the username for an unlinked InstitutionAccount
    and return a tuple containing the following:
      0: Account Type (e.g. 'cas')
      1: Institution Slug
      2: User ID for the institution
    """
    username_split = username.split("-")
    if len(username_split) < 3:
        raise ValueError("Value passed to get_account_username_split " +
                "was not the username for an unlinked InstitutionAccount.")
    slug = "-".join(username_split[1:len(username_split)-1])
    return (username_split[0], slug, username_split[len(username_split)-1])


def get_input(prompt):
    """
    Forwards to either raw_input or input, depending on Python version
    """
    try:
        return raw_input(prompt)
    except NameError:
        return input(prompt)


def get_protocol(request):
    """
    Returns the protocol request is using ('http' | 'https')
    """
    return 'https' if request.is_secure() else 'http'


def get_random_username():
    """
    Returns a username generated from current timestamp + random string
    """
    return "tmp-%s_%s" % (timezone.now().strftime("%Y%m%d%H%M%S%f"),
            get_random_string(5))


def get_redirect_url(request, use_referer=False, default_url=None):
    """
    Returns the URL to redirect to once business at the current
    URL is completed.

    Picks the first usable URL from the following list:
      1. URL provided as GET parameter under REDIRECT_FIELD_NAME
      2. Referring page if use_referer is True, and set in header
      3. default_url parameter
      4. UNIAUTH_LOGIN_REDIRECT_URL setting
    """
    redirect_url = request.GET.get(REDIRECT_FIELD_NAME)
    if not redirect_url:
        if use_referer:
            redirect_url = request.META.get('HTTP_REFERER')
        if not redirect_url:
            redirect_url = resolve_url(default_url or
                    get_setting('UNIAUTH_LOGIN_REDIRECT_URL'))
        prefix = urlunparse(
                (get_protocol(request), request.get_host(), '', '', '', ''),
        )
        if redirect_url.startswith(prefix):
            redirect_url = redirect_url[len(prefix):]
    return redirect_url


def get_service_url(request, redirect_url=None):
    """
    Returns the service URL to provide to the CAS
    server for the provided request.

    Accepts an optional redirect_url, which defaults
    to the value of get_redirect_url(request).
    """
    service_url = urlunparse(
            (get_protocol(request), request.get_host(),
            request.path, '', '', ''),
    )
    query_params = request.GET.copy()
    query_params[REDIRECT_FIELD_NAME] = redirect_url or \
            get_redirect_url(request)
    # The CAS server may have added the ticket as an extra query
    # parameter upon checking the credentials - ensure it is ignored
    query_params.pop('ticket', None)
    service_url += '?' + urlencode(query_params)
    return service_url


def get_setting(setting_name):
    """
    Returns the value of the setting with the provided name
    if set. Returns the value in DEFAULT_SETTING_VALUES
    otherwise.
    """
    try:
        return getattr(settings, setting_name)
    except AttributeError:
        return DEFAULT_SETTING_VALUES[setting_name]


def is_tmp_user(user):
    """
    Returns whether the provided user is a temporary one:

    By default, this includes users midway through verifying
    their profile (have usernames starting with "tmp").

    If the UNIAUTH_ALLOW_STANDALONE_ACCOUNTS setting is
    False, it includes Institution Account logins (such as
    CAS) that have not been linked to a Uniauth profile yet.
    """
    return user.username and (user.username.startswith('tmp-') or \
            (not get_setting('UNIAUTH_ALLOW_STANDALONE_ACCOUNTS') and \
                    is_unlinked_account(user)))


def is_unlinked_account(user):
    """
    Returns whether the provided user has authenticated via
    an InstitutionAccount not yet linked to a Uniauth profile.
    """
    return user.username and user.username.startswith('cas-')



# Generated by Django 2.1.4 on 2019-01-25 17:51

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion

class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Institution',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30)),
                ('slug', models.CharField(max_length=30, unique=True)),
                ('cas_server_url', models.URLField()),
            ],
        ),
        migrations.CreateModel(
            name='InstitutionAccount',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('cas_id', models.CharField(max_length=30)),
                ('institution', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accounts', to='uniauth.Institution')),
            ],
        ),
        migrations.CreateModel(
            name='LinkedEmail',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('address', models.EmailField(max_length=254)),
                ('is_verified', models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name='UserProfile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='linkedemail',
            name='profile',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='linked_emails', to='uniauth.UserProfile'),
        ),
        migrations.AddField(
            model_name='institutionaccount',
            name='profile',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='accounts', to='uniauth.UserProfile'),
        ),
    ]



from datetime import timedelta
from django.conf import settings
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.db import transaction
from django.db.utils import IntegrityError
from django.test import override_settings, TestCase
from django.utils import timezone
from uniauth.models import Institution, InstitutionAccount, LinkedEmail, \
        UserProfile


class ModelSignalTests(TestCase):
    """
    Tests that the custom model signals fire and
    behave as expected
    """

    def test_create_user_profile_signal(self):
        """
        Ensure a UserProfile is automatically created
        whenever Users are
        """
        user = User.objects.create(username="new-user")
        self.assertTrue(UserProfile.objects.filter(
                user__username="new-user").exists())
        user2 = User.objects.create(username="testusr@example.com",
                email="test.user@example.com")
        self.assertTrue(UserProfile.objects.filter(
                user__email="test.user@example.com").exists())

    @override_settings(PASSWORD_RESET_TIMEOUT_DAYS=4)
    def test_clear_old_tmp_users_signal(self):
        """
        Ensure old temporary users are deleted whenever
        a new User is created
        """
        User.objects.all().delete()
        User.objects.create(username="not-temporary-user")
        for i in range(10):
            User.objects.create(username="tmp-%d-days-ago"%i)
        # We must update the date_joined in a different for loop,
        # because otherwise, the users could get deleted on the
        # create signal we're trying to test!
        for i in range(10):
            date_joined = timezone.now() - timedelta(days=i)
            user = User.objects.get(username="tmp-%d-days-ago"%i)
            user.date_joined = date_joined
            user.save()
        # Create another object to (hopefully) trigger the tmp
        # user deletion signal
        User.objects.create(username="another-user")

        expected_num_users = 10 - (settings.PASSWORD_RESET_TIMEOUT_DAYS + 1) + 2
        self.assertEqual(User.objects.count(), expected_num_users)
        self.assertTrue(User.objects.filter(username="not-temporary-user")\
                .exists())
        self.assertTrue(User.objects.filter(username="another-user").exists())
        for i in range(settings.PASSWORD_RESET_TIMEOUT_DAYS + 1):
            self.assertTrue(User.objects.filter(username="tmp-%d-days-ago"%i)\
                    .exists())
        for i in range(settings.PASSWORD_RESET_TIMEOUT_DAYS + 1, 10):
            self.assertFalse(User.objects.filter(username="tmp-%d-days-ago"%i)\
                    .exists())


class UserProfileModelTests(TestCase):
    """
    Tests the UserProfile model in models.py
    """

    def test_user_profile_model_basic(self):
        """
        Ensure the basic model methods + attributes work properly
        """
        user = User.objects.create(username="new-user")
        self.assertNotEqual(user.uniauth_profile, None)
        self.assertEqual(user.uniauth_profile.user, user)
        self.assertEqual(str(user.uniauth_profile), "new-user")
        user2 = User.objects.create(username="other-user",
                email="other@gmail.com")
        self.assertEqual(str(user2.uniauth_profile), "other@gmail.com")
        result = UserProfile.objects.get(user__username="new-user")
        self.assertEqual(result, user.uniauth_profile)
        with self.assertRaises(IntegrityError):
            UserProfile.objects.create()

    def test_user_profile_get_display_id(self):
        """
        Ensure the get_display_id method works properly
        """
        user = User.objects.create(username="new-user")
        self.assertEqual(user.uniauth_profile.get_display_id(), "new-user")
        user = User.objects.create(username="cas-example-inst-id123")
        self.assertEqual(user.uniauth_profile.get_display_id(), "id123")
        user = User.objects.create(username="john.doe@example.com",
                email="otheraddress@example.com")
        self.assertEqual(user.uniauth_profile.get_display_id(), "john.doe")


class LinkedEmailModelTests(TestCase):
    """
    Tests the LinkedEmail model in models.py
    """

    def test_linked_email_model_basic(self):
        """
        Ensure the basic model methods + attributes work properly
        """
        user = User.objects.create(username="new-user")
        linked_email = LinkedEmail.objects.create(profile=user.uniauth_profile,
                address="example@gmail.com", is_verified=True)
        self.assertEqual(linked_email.profile, user.uniauth_profile)
        self.assertEqual(linked_email.address, "example@gmail.com")
        self.assertTrue(linked_email.is_verified)
        self.assertTrue("new-user" in str(linked_email))
        self.assertTrue("example@gmail.com" in str(linked_email))
        result = LinkedEmail.objects.get(profile=user.uniauth_profile,
                address="example@gmail.com")
        self.assertEqual(result, linked_email)

    def test_linked_email_model_clean(self):
        """
        Ensure the model prevents saving invalid states
        """
        user1 = User.objects.create(username="new-user-1")
        user2 = User.objects.create(username="new-user-2")
        user3 = User.objects.create(username="new-user-3")

        # Don't allow shared emails if setting is False
        with self.settings(UNIAUTH_ALLOW_SHARED_EMAILS=False):
            LinkedEmail.objects.create(profile=user1.uniauth_profile,
                    address="shared@example.com", is_verified=True)
            try:
                email = LinkedEmail(profile=user2.uniauth_profile,
                        address="shared@example.com", is_verified=True)
                email.clean()
                email.save()
                self.fail("Adding shared email address succeeded when "
                        "UNIAUTH_ALLOW_SHARED_EMAILS was False.")
            except ValidationError:
                pass
            self.assertFalse(LinkedEmail.objects.filter(
                    profile=user2.uniauth_profile,
                    address="shared@example.com").exists())

        # Don't allow more than the max number of linked emails per profile
        with self.settings(UNIAUTH_MAX_LINKED_EMAILS=5):
            for i in range(settings.UNIAUTH_MAX_LINKED_EMAILS):
                LinkedEmail.objects.create(profile=user3.uniauth_profile,
                        address="email%d@example.com"%i, is_verified=(i%2==0))
            try:
                email = LinkedEmail(profile=user3.uniauth_profile,
                        address="another@example.com")
                email.clean()
                email.save()
                self.fail("Able to link more than UNIAUTH_MAX_LINKED_EMAILS")
            except ValidationError:
                pass
            self.assertFalse(LinkedEmail.objects.filter(profile=user3.uniauth_profile,
                    address="another@example.com").exists())

        # Allow any number of linked emails if setting is <= 0
        with self.settings(UNIAUTH_MAX_LINKED_EMAILS=-1):
            for i in range(100):
                email, _ = LinkedEmail.objects.get_or_create(
                        profile=user3.uniauth_profile,
                        address="email%d@example.com"%i, is_verified=(i%2==0))
                email.clean()
                email.save()
            self.assertEqual(LinkedEmail.objects.filter(
                    profile=user3.uniauth_profile).count(), 100)


class InstitutionModelTests(TestCase):
    """
    Tests the Institution model in models.py
    """

    def test_institution_model_basic(self):
        """
        Ensure the basic model methods + attributes work properly
        """
        institution = Institution.objects.create(name="Test Inst",
                slug="test-inst", cas_server_url="https://fed.example.edu/")
        self.assertEqual(institution.name, "Test Inst")
        self.assertEqual(institution.slug, "test-inst")
        self.assertEqual(institution.cas_server_url, "https://fed.example.edu/")
        self.assertEqual(str(institution), "test-inst")
        result = Institution.objects.get(slug="test-inst")
        self.assertEqual(result, institution)
        with self.assertRaises(IntegrityError):
            Institution.objects.create(name="Test-inst", slug="test-inst",
                    cas_server_url="https://www.different.edu/")


class InstitutionAccountModelTests(TestCase):
    """
    Tests the InstitutionAccount model in models.py
    """

    def test_institution_account_model_basic(self):
        """
        Ensure the basic model methods + attributes work properly
        """
        user = User.objects.create(username="new-user")
        institution = Institution.objects.create(name="Test Inst",
                slug="test-inst", cas_server_url="https://fed.example.edu/")
        account = InstitutionAccount.objects.create(
                profile=user.uniauth_profile,
                institution=institution, cas_id="netid123")
        self.assertEqual(account.profile, user.uniauth_profile)
        self.assertEqual(account.institution, institution)
        self.assertEqual(account.cas_id, "netid123")
        self.assertTrue("test-inst" in str(account))
        self.assertTrue("new-user" in str(account))
        result = InstitutionAccount.objects.get(institution=institution,
                cas_id="netid123")
        self.assertEqual(result, account)

    def test_institution_account_uniqueness_constraint(self):
        """
        Ensure creating multiple InstitutionAccounts with the same
        institution and cas_id fails
        """
        user1 = User.objects.create(username="new-user")
        user2 = User.objects.create(username="other-user")
        institution = Institution.objects.create(name="Test Inst",
                slug="test-inst", cas_server_url="https://fed.example.edu/")
        cas_id = "netid123"
        InstitutionAccount.objects.create(
                profile=user1.uniauth_profile,
                institution=institution, cas_id=cas_id)
        with self.assertRaises(IntegrityError):
            InstitutionAccount.objects.create(
                    profile=user2.uniauth_profile,
                    institution=institution, cas_id=cas_id)
